"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClassStoresPlugin = void 0;
const fs_jetpack_1 = __importDefault(require("fs-jetpack"));
const path_1 = __importDefault(require("path"));
const Generator_1 = require("./Generator");
const Logger_1 = require("./Logger");
const Types_1 = require("./Types");
const pluginConfig = Object.assign({ projectRoot: null, projectRootDirectory: null, storesPath: null, storesDirectory: null, storesFileName: null }, Types_1.defaultPluginConfig);
const storeNameRegex = /(Store|store)\.ts$/;
function ClassStoresPlugin(configuration) {
    for (let configurationKey in configuration) {
        if (configuration[configurationKey]) {
            pluginConfig[configurationKey] = configuration[configurationKey];
        }
    }
    let viteConfig;
    let server;
    const context = new Generator_1.Context();
    return {
        name: 'class-stores-loader',
        configResolved(config) {
            return __awaiter(this, void 0, void 0, function* () {
                pluginConfig.projectRoot = config.root;
                pluginConfig.storesPath = path_1.default.resolve(pluginConfig.projectRoot, pluginConfig.storesPath);
                pluginConfig.storesDirectory = fs_jetpack_1.default.cwd(pluginConfig.storesPath);
                pluginConfig.projectRootDirectory = fs_jetpack_1.default.cwd(pluginConfig.projectRoot);
                viteConfig = config;
                context.init(pluginConfig);
                context.process();
            });
        },
        configureServer(devServer) {
            server = devServer;
            server.watcher.on('all', (eventName, filePath, stats) => {
                if (!['add', 'change', 'unlink'].includes(eventName)) {
                    return;
                }
                if (isGeneratedFilePath(filePath)) {
                    return;
                }
                if (!filePath.startsWith(pluginConfig.storesPath)) {
                    return;
                }
                context.reloadModules();
                if (['change', 'unlink'].includes(eventName)) {
                    const store = context.getStoreByFilePath(filePath);
                    if (!store) {
                        return;
                    }
                    eventName === 'unlink'
                        ? (0, Logger_1.infoLog)(`Store file deletion detected: ${store.className}`)
                        : (0, Logger_1.infoLog)(`Store file change detected: ${store.className}`);
                    (0, Logger_1.infoLog)(`Rebuilding stores...`);
                    context.rebuild();
                    return;
                }
                let store = context.getStoreByFilePath(filePath);
                if (!store) {
                    (0, Logger_1.infoLog)(`New file detected: ${filePath}`);
                    (0, Logger_1.infoLog)(`Rebuilding store meta...`);
                    context.process();
                }
                // Just to ensure it was actually a new store, not a random file
                store = context.getStoreByFilePath(filePath);
                if (store) {
                    (0, Logger_1.infoLog)(`Addition was a new store: ${store.className}`);
                    (0, Logger_1.infoLog)(`Rebuilding stores...`);
                    context.rebuild();
                }
            });
        },
        buildStart(options) {
            context.writeFiles();
        },
    };
}
exports.ClassStoresPlugin = ClassStoresPlugin;
/**
 * Checks if the specified file is one that's generated by the plugin.
 * This is used within things like the HMR handler to prevent infinite loops.
 * @param {string} file
 */
function isGeneratedFilePath(file) {
    const paths = [
        pluginConfig.storesDirectory.path(pluginConfig.storesFileName),
    ];
    return paths.includes(file);
}
//# sourceMappingURL=ClassStoresPlugin.js.map