"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.StoreMeta = exports.Context = void 0;
const chalk_1 = __importDefault(require("chalk"));
const fs_jetpack_1 = __importDefault(require("fs-jetpack"));
const lodash_1 = __importDefault(require("lodash"));
const path = __importStar(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const crypto_1 = __importDefault(require("crypto"));
const Common_1 = require("../Common");
const Imports_1 = require("./Builders/Imports");
const Object_1 = require("./Builders/Object");
const StoreLoader_1 = require("./Builders/StoreLoader");
const Logger_1 = require("./Logger");
const { factory } = typescript_1.default;
const GENERATED_FILE_DIR = 'Generated';
class Context {
    constructor() {
        Object.defineProperty(this, "program", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * File paths to all files in the project(loaded by typescript compiler)
         * @type {string[]}
         */
        Object.defineProperty(this, "filePaths", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        /**
         * File paths to all of our store files in the project
         * @type {string[]}
         */
        Object.defineProperty(this, "storeFilePaths", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * When we first load our stores, we'll take a hash of the content
         * Later on this hash will be used when there is a change, mainly to
         * determine if we need to do some file writes, when writing stores.d.ts,
         * even if contents are the same, this makes vite do a full page reload...
         * @type {Map<string, string>}
         */
        Object.defineProperty(this, "storeFileHashes", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        /**
         * All of the store modules in our project
         *
         * @type {{module: ts.SourceFile, store: StoreMeta}[]}
         */
        Object.defineProperty(this, "modules", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "stores", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "config", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "compilerHost", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "printer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "typeChecker", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "parsedCommandLine", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * This is the directory where all of our generated files will be stored
         * @type {FSJetpack}
         * @private
         */
        Object.defineProperty(this, "generatedDir", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
    }
    init(pluginConfig) {
        Context.instance = this;
        this.config = pluginConfig;
        const configPath = typescript_1.default.findConfigFile(pluginConfig.projectRoot, typescript_1.default.sys.fileExists, "tsconfig.json");
        if (!configPath) {
            throw new Error("Could not find a valid 'tsconfig.json'.");
        }
        this.parsedCommandLine = typescript_1.default.parseJsonConfigFileContent(typescript_1.default.readConfigFile(configPath, typescript_1.default.sys.readFile).config, typescript_1.default.sys, "./");
        this.generatedDir = fs_jetpack_1.default.dir(this.config.storesDirectory.path(GENERATED_FILE_DIR));
        this.setFilePaths();
        this.compilerHost = typescript_1.default.createCompilerHost(this.parsedCommandLine.options);
        this.program = typescript_1.default.createProgram(this.filePaths, this.parsedCommandLine.options, this.compilerHost);
        this.printer = typescript_1.default.createPrinter({ newLine: typescript_1.default.NewLineKind.LineFeed });
        this.typeChecker = this.program.getTypeChecker();
        this.reloadModules(false);
        this.loadGeneratedFileHash(pluginConfig.storesFileName);
        this.loadGeneratedFileHash(pluginConfig.storeLoaderFile);
    }
    /**
     * Set the store file paths & all project file paths
     *
     * @returns {string[]}
     */
    setFilePaths() {
        const paths = this.config.storesDirectory.find({
            directories: false,
            files: true,
            recursive: true,
            matching: ['*Store.ts', '*store.ts'],
        }).map(filePath => typescript_1.default.sys.resolvePath(this.config.storesDirectory.path(filePath)));
        this.storeFilePaths = paths;
        return this.filePaths = lodash_1.default.uniq(paths.concat(...this.parsedCommandLine.fileNames.map(f => typescript_1.default.sys.resolvePath(f))));
    }
    /**
     * Load all of our store modules and create a "store meta" class.
     */
    reloadModules(reloadFilePaths = true) {
        if (reloadFilePaths) {
            this.setFilePaths();
        }
        this.modules = this.program.getSourceFiles()
            .filter(module => this.storeFilePaths.includes(module.fileName))
            .map(module => ({ module, store: new StoreMeta(module, this.config.storesPath) }));
        for (let module of this.modules) {
            this.addFileHash(module.module.fileName, module.module.getFullText());
        }
    }
    addFileHash(filePath, content) {
        const hash = crypto_1.default.createHash('sha256').update(content).digest('hex');
        if (this.storeFileHashes.has(filePath)) {
            const oldHash = this.storeFileHashes.get(filePath);
            if (oldHash === hash) {
                return false;
            }
            this.storeFileHashes.set(filePath, hash);
            (0, Logger_1.warnLog)("Store file hash is different: " + filePath);
            return true;
        }
        this.storeFileHashes.set(filePath, hash);
        return true;
    }
    loadGeneratedFileHash(file) {
        if (this.generatedDir.exists(file)) {
            const content = this.generatedDir.read(file);
            this.addFileHash(this.generatedDir.path(file), content);
        }
    }
    /**
     * MAKE SURE TO PASS A FULL ABSOLUTE PATH TO THIS METHOD!
     *
     * @param {string} filePath
     * @returns {boolean}
     * @private
     */
    fileHasChanged(filePath) {
        const fPath = fs_jetpack_1.default.path(filePath);
        if (!this.storeFileHashes.has(fPath)) {
            return true;
        }
        const content = fs_jetpack_1.default.read(fPath);
        const hash = crypto_1.default.createHash('sha256').update(content).digest('hex');
        return this.storeFileHashes.get(filePath) !== hash;
    }
    process() {
        this.stores = [];
        for (const { module, store } of this.modules) {
            const outStore = this.processModule(module, store);
            if (!outStore)
                continue;
            this.stores.push(outStore.finalize());
        }
    }
    writeFiles() {
        this.writeStoreDeclaration();
    }
    writeStoreDeclaration() {
        var _a;
        const storeCount = (_a = this.stores) === null || _a === void 0 ? void 0 : _a.length;
        if (!storeCount) {
            (0, Logger_1.warnLog)('No stores located... skipping code-gen.');
            return;
        }
        (0, Logger_1.infoLog)('Located', chalk_1.default.reset(storeCount), 'stores:');
        (0, Logger_1.basicLog)(this.stores.map(s => "   - " + chalk_1.default.reset(s.className)).join('\n'));
        const vueDtsFile = createVueDtsFile(this.stores);
        const { didWrite, filePath } = this.writeTsFile(this.generatedDir.path(this.config.storesFileName), vueDtsFile);
        if (didWrite) {
            (0, Logger_1.successLog)('Generated stores vue declaration file at:', chalk_1.default.reset(this.generatedDir.path(filePath).replace(this.config.projectRoot, '')));
        }
        const loaderResult = this.writeTsFile(this.generatedDir.path(this.config.storeLoaderFile), (0, StoreLoader_1.createStoreLoaderModule)(this.stores));
        if (loaderResult.didWrite) {
            (0, Logger_1.successLog)('Generated store loader file at:', chalk_1.default.reset(this.generatedDir.path(loaderResult.filePath).replace(this.config.projectRoot, '')));
        }
    }
    rebuild() {
        this.process();
        this.writeFiles();
    }
    processModule(module, store) {
        var _a, _b, _c, _d, _e;
        for (let statement of module.statements) {
            // Search for our `export myStoreName = new MyStore();` statement and extract `myStoreName`
            if (typescript_1.default.isVariableStatement(statement)) {
                if (((_a = statement.modifiers) === null || _a === void 0 ? void 0 : _a.some(m => m.kind === typescript_1.default.SyntaxKind.ExportKeyword)) && ((_c = (_b = statement === null || statement === void 0 ? void 0 : statement.declarationList) === null || _b === void 0 ? void 0 : _b.declarations) === null || _c === void 0 ? void 0 : _c.length) === 1) {
                    const declaration = statement.declarationList.declarations[0];
                    if (typescript_1.default.isIdentifier(declaration.name)) {
                        store.exportName = declaration.name.text;
                    }
                }
            }
            // Search for our `export class MyStore extends Store<MyStore, IMyStoreState> {` statement and extract `MyStore`
            if (typescript_1.default.isClassDeclaration(statement)) {
                // Ensure our class extends Store
                const extendsStore = (_d = statement.heritageClauses) === null || _d === void 0 ? void 0 : _d.some(h => {
                    if (h.token !== typescript_1.default.SyntaxKind.ExtendsKeyword)
                        return false;
                    return h.types.some(t => {
                        var _a;
                        if (!typescript_1.default.isExpressionWithTypeArguments(t))
                            return false;
                        if (!typescript_1.default.isCallExpression(t.expression))
                            return false;
                        const callExpression = t.expression;
                        if (!typescript_1.default.isIdentifier(callExpression.expression))
                            return false;
                        return ((_a = callExpression.expression.text) === null || _a === void 0 ? void 0 : _a.toLowerCase()) === "store";
                    });
                });
                if (extendsStore && typescript_1.default.isIdentifier(statement.name)) {
                    store.className = statement.name.text;
                    // Now we have to look for the `public static vueBinding = 'x';` statement and pull out it's value.
                    for (let member of statement.members) {
                        if (typescript_1.default.isPropertyDeclaration(member)) {
                            if (!((_e = member.modifiers) === null || _e === void 0 ? void 0 : _e.some(m => m.kind === typescript_1.default.SyntaxKind.StaticKeyword)))
                                continue;
                            if (typescript_1.default.isIdentifier(member.name) && member.name.text === "vueBinding") {
                                if (typescript_1.default.isStringLiteral(member.initializer)) {
                                    store.vueBinding = member.initializer.text;
                                    break;
                                }
                            }
                        }
                    }
                    const actions = [];
                    for (let member of statement.members) {
                        if (typescript_1.default.isMethodDeclaration(member)) {
                            if (typescript_1.default.isIdentifier(member.name)) {
                                const action = new ActionMeta(member, member.name.text);
                                actions.push(action);
                            }
                        }
                    }
                    store.actions = actions;
                }
            }
        }
        return store.isValid() ? store : undefined;
    }
    isStoreFile(path) {
        return this.stores.some(s => s.absFilePath === path);
    }
    getStoreByFilePath(path) {
        return this.stores.find(s => s.absFilePath === path);
    }
    writeTsFile(filePath, nodes) {
        const result = { didWrite: false, filePath: filePath };
        const sourceFile = typescript_1.default.createSourceFile(filePath, '', this.parsedCommandLine.options.target, false, typescript_1.default.ScriptKind.TS);
        const source = this.printer.printList(typescript_1.default.ListFormat.MultiLine, factory.createNodeArray(nodes), sourceFile);
        result.didWrite = this.writeFile(filePath, source);
        return result;
    }
    writeFile(filePath, contents) {
        if (!this.addFileHash(filePath, contents)) {
            return false;
        }
        fs_jetpack_1.default.write(filePath, contents);
        return true;
    }
    static printNode(node) {
        try {
            const sourceFile = typescript_1.default.createSourceFile('temp.ts', '', this.instance.parsedCommandLine.options.target, false, typescript_1.default.ScriptKind.TS);
            (0, Logger_1.infoLog)('Printed node > \n' + this.instance.printer.printNode(typescript_1.default.EmitHint.Unspecified, node, sourceFile));
        }
        catch (error) {
            (0, Logger_1.errorLog)("Failed to print node", error);
        }
    }
}
exports.Context = Context;
class StoreMeta {
    constructor(file, absStoresPath) {
        /**
         * The absolute path to the store file
         */
        Object.defineProperty(this, "absFilePath", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The relative path of the store file, for ex, if importing in /src/stores/, and our file is at /src/stores/MyStore.ts, this will be MyStore.ts
         */
        Object.defineProperty(this, "relStoreFilePath", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The relative import path from Stores/.generated/StoreLoader.ts to the store location
         * @type {string}
         */
        Object.defineProperty(this, "loaderImportPath", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The name of the store file
         */
        Object.defineProperty(this, "name", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * The name of the class that extends Store
         */
        Object.defineProperty(this, "className", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * This will hold the name of our store export `export const store = new MyStore();`
         */
        Object.defineProperty(this, "exportName", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * If the dev defined `public static vueBinding = 'x';` in their store, this will hold that value
         */
        Object.defineProperty(this, "vueBinding", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "actions", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        this.absFilePath = file.fileName;
        this.name = path.basename(file.fileName);
        this.relStoreFilePath = path.relative(Context.instance.generatedDir.path(), file.fileName);
        this.loaderImportPath = this.relStoreFilePath; // path.relative(file.fileName, Context.instance.generatedDir.path(Context.instance.config.storeLoaderFile));
    }
    isValid() {
        return this.className !== undefined && this.exportName !== undefined;
    }
    finalize() {
        this.formatVueBinding();
        return this;
    }
    formatVueBinding() {
        if (this.vueBinding) {
            return;
        }
        this.vueBinding = (0, Common_1.formatVueBindingName)(this.vueBinding, this.className);
    }
    metaObject(includeImport = true) {
        return {
            className: this.className,
            importPath: this.loaderImportPath,
            exportName: this.exportName,
            vueBinding: this.vueBinding,
            actions: JSON.parse(JSON.stringify(this.actions)),
            module: (0, Object_1.unwrappableNode)((0, Imports_1.createLazyImportGlobNode)(this.loaderImportPath))
        };
    }
}
exports.StoreMeta = StoreMeta;
class ActionMeta {
    constructor(method, name) {
        Object.defineProperty(this, "name", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "params", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "method", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "signature", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.method = method;
        this.name = name;
        this.signature = Context.instance.typeChecker.getSignatureFromDeclaration(this.method);
        for (let parameter of this.signature.parameters) {
            const ptype = Context.instance.typeChecker.typeToString(Context.instance.typeChecker.getTypeAtLocation(parameter.valueDeclaration));
            let defaultValue = undefined;
            if (typescript_1.default.isParameter(parameter.valueDeclaration)) {
                if (parameter.valueDeclaration.initializer) {
                    defaultValue = parameter.valueDeclaration.initializer.getText();
                }
            }
            this.params.push({
                name: parameter.name,
                type: ptype,
                defaultValue,
            });
        }
    }
    toJSON() {
        return {
            name: this.name,
            params: this.params
        };
    }
}
function createVueDtsFile(stores) {
    /**
     * This code will generate the following:
     *
     * declare module "@vue/runtime-core" {
     *   import {storeExportName} from "./StoreName";
     * 	 interface ComponentCustomProperties {
     * 		$storeVueBinding: typeof storeExportName;
     * 	 }
     * }
     *
     * export {};
     */
    /**
     * This will generate:
     * $storeVueBinding: typeof storeExportName;
     */
    const componentProperties = stores.map(store => factory.createPropertySignature(undefined, factory.createIdentifier(store.vueBinding), undefined, factory.createTypeQueryNode(factory.createIdentifier(store.exportName), undefined)));
    /**
     * This will generate our imports:
     * import {storeExportName} from "./StoreName";
     */
    const storeImports = stores.map(store => (0, Imports_1.createRelativeImportNode)([store.exportName], store.relStoreFilePath));
    /**
     * This will generate the main declare module/interface block
     */
    const declareModuleBlock = factory.createModuleDeclaration([factory.createModifier(typescript_1.default.SyntaxKind.DeclareKeyword)], factory.createStringLiteral("@vue/runtime-core"), factory.createModuleBlock([
        ...storeImports,
        factory.createInterfaceDeclaration(undefined, factory.createIdentifier("ComponentCustomProperties"), undefined, undefined, componentProperties)
    ]), typescript_1.default.NodeFlags.ExportContext /*| ts.NodeFlags.Ambient*/ | typescript_1.default.NodeFlags.ContextFlags);
    return [
        declareModuleBlock,
        factory.createExportDeclaration(undefined, false, factory.createNamedExports([]), undefined, undefined),
    ];
}
//# sourceMappingURL=Generator.js.map